#include "CASACLE.hpp"
#include "FAST/Data/Image.hpp"
#include "FAST/Data/Segmentation.hpp"
#include "FAST/Data/LineSet.hpp"
#include "FAST/Algorithms/GaussianSmoothingFilter/GaussianSmoothingFilter.hpp"
#include "FAST/Algorithms/GradientVectorFlow/EulerGradientVectorFlow.hpp"
#include "FAST/Algorithms/GradientVectorFlow/MultigridGradientVectorFlow.hpp"
#include "FAST/Algorithms/TubeSegmentationAndCenterlineExtraction/InverseGradientSegmentation.hpp"
#include "RTCE.hpp"
#include <stack>
#include <iostream>
#include <fstream>
#include <math.h>


namespace fast {

    CASACLE::CASACLE() {

        createInputPort<Image>(0);
        createOutputPort<Segmentation>(0, OUTPUT_DEPENDS_ON_INPUT, 0);
        createOutputPort<LineSet>(1, OUTPUT_DEPENDS_ON_INPUT, 0);
        createOutputPort<Image>(2, OUTPUT_DEPENDS_ON_INPUT, 0);

        createOpenCLProgram(std::string(FAST_SOURCE_DIR) + "Algorithms/TubeSegmentationAndCenterlineExtraction/TubeSegmentationAndCenterlineExtraction.cl");

        mSensitivity = 0.5;
        mMinimumRadius = 1;
        mMaximumRadius = 5;
        mRadiusStep = 0.5;
        //mExtractDarkStructures = false;
        mExtractDarkStructures = true;
        mMinimumIntensity = -std::numeric_limits<float>::max();
        mMaximumIntensity = std::numeric_limits<float>::max();
        // Blur has to be adapted to noise level in image
        //mStDevBlurSmall = 0.5;
        mStDevBlurSmall = 1.0;
        mStDevBlurLarge = 1.0;
        gaussianStrength = 1.0;
        croppingFraction = 0.135;
        croppingEnabled = true;
        gaussianEnabled = false;
    }

    void CASACLE::loadPreset() {
    }

    void CASACLE::setMinimumRadius(
        float radius) {
        mMinimumRadius = radius;
    }

    void CASACLE::setMaximumRadius(
        float radius) {
        mMaximumRadius = radius;
    }

    void CASACLE::setRadiusStep(float step) {
        mRadiusStep = step;
    }

    void CASACLE::setSensitivity(
        float sensitivity) {
        mSensitivity = sensitivity;
        if (mSensitivity == 0) {
            mSensitivity = 0.001;
        }
        else if (mSensitivity == 1) {
            mSensitivity = 0.999;
        }
    }

    void CASACLE::setMinimumIntensity(
        float intensity) {
        mMinimumIntensity = intensity;
    }

    void CASACLE::setMaximumIntensity(
        float intensity) {
        mMaximumIntensity = intensity;
    }

    void CASACLE::extractDarkTubes() {
        mExtractDarkStructures = true;
    }

    void CASACLE::enableGaussian(){
        gaussianEnabled = true;
    }

    void CASACLE::setGaussianStrength(float strength){
        gaussianStrength = strength;
    }

    void CASACLE::enableCropping(){
        croppingEnabled = true;
    }

    void CASACLE::setCroppingFraction(float fraction){
        croppingFraction = fraction;
    }

    void CASACLE::extractBrightTubes() {
        mExtractDarkStructures = false;
    }

    ProcessObjectPort CASACLE::getSegmentationOutputPort() {
        return getOutputPort(0);
    }

    ProcessObjectPort CASACLE::getCenterlineOutputPort() {
        return getOutputPort(1);
    }

    ProcessObjectPort CASACLE::getTDFOutputPort() {
        return getOutputPort(2);
    }

    void CASACLE::execute(){

        Image::pointer input = getStaticInputData<Image>();
        Vector3f spacing = input->getSpacing();
        Vector3ui size = input->getSize();
        float smallestSpacing = spacing.minCoeff();
        float largestSpacing = spacing.maxCoeff();

        Image::pointer gradients;
        Image::pointer largeTDF;
        Image::pointer largeRadius;
        Image::pointer GVFfield;
        Image::pointer smoothedImage = input;
        //Add smoothing + nlm options here

        if (croppingEnabled){
            input = performCropping(input,croppingFraction);
            Image::pointer smoothedImage = input;
            std::cout << "cropping done" << std::endl;
        }

        if (gaussianEnabled){
            GaussianSmoothingFilter::pointer gaussianFilter = GaussianSmoothingFilter::New();
            gaussianFilter->setInputData(input);
            gaussianFilter->setStandardDeviation(gaussianStrength);
            //filter->setMaskSize(7);
            gaussianFilter->setOutputType(TYPE_FLOAT);
            gaussianFilter->update();
            smoothedImage = gaussianFilter->getOutputData<Image>();
            smoothedImage->setSpacing(spacing);
            std::cout << "gaus done" << std::endl;
        }
        
        GVFfield = createGradients(smoothedImage);
        std::cout << "gradients done" << std::endl;
        gradients = GVFfield;
        GVFfield = runGradientVectorFlow(GVFfield);
        std::cout << "GVF done" << std::endl;
        std::cout << "Large TDF starting" << std::endl;
        runCentralTubeDetectionFilter(GVFfield, 0.5, 0.5, 100, largeTDF, largeRadius, 1, mMinimumRadius, mMaximumRadius);
        std::cout << "large TDF done" << std::endl;
        Image::pointer smallTDF;
        Image::pointer smallRadius;
        //gradients = createGradients(smoothedImage);
        runCentralTubeDetectionFilter(gradients, 0.5, 0.5, 100, smallTDF, smallRadius, 0, mMinimumRadius, mMaximumRadius);
        std::cout << "small TDF done" << std::endl;
        RTCE::pointer centerlineExtraction = RTCE::New();

        LineSet::pointer centerline;

        centerlineExtraction->setInputData(0, largeTDF);
        centerlineExtraction->setInputData(1, GVFfield);
        centerlineExtraction->setInputData(2, largeRadius);
        centerlineExtraction->setInputData(3, smallTDF);
        centerlineExtraction->setInputData(4, gradients);
        centerlineExtraction->setInputData(5, smallRadius);
        centerlineExtraction->setInputData(6, input);
        centerlineExtraction->update();
        centerline = centerlineExtraction->getOutputData<LineSet>();

        InverseGradientSegmentation::pointer segmentation = InverseGradientSegmentation::New();
        segmentation->setInputConnection(centerlineExtraction->getOutputPort(1));
        segmentation->setInputData(1, GVFfield);
        segmentation->update();
        Segmentation::pointer segmentationVolume = segmentation->getOutputData<Segmentation>();
        setStaticOutputData<Segmentation>(0, segmentationVolume);
        setStaticOutputData<LineSet>(1, centerline);
        Image::pointer TDF;
        setStaticOutputData<Image>(2, largeTDF);

    }

    Image::pointer CASACLE::performCropping(Image::pointer input, float croppingFraction){
        float fraction = croppingFraction;
        uint startX = round(input->getWidth()*fraction);
        uint sizeX = input->getWidth() - round(input->getWidth()*fraction * 2);
        uint startY = round(input->getHeight()*fraction);
        uint sizeY = input->getHeight() - round(input->getHeight()*fraction * 2);
        uint startZ = round(input->getDepth()*fraction);
        uint sizeZ = input->getDepth() - round(input->getDepth()*fraction * 2);

        // Make size dividable by 4
        sizeX = sizeX + (4 - sizeX % 4);
        sizeY = sizeY + (4 - sizeY % 4);
        sizeZ = sizeZ + (4 - sizeZ % 4);

        Image::pointer output = input->crop(Vector3ui(startX, startY, startZ), Vector3ui(sizeX, sizeY, sizeZ));
        return output;
    }

    Image::pointer CASACLE::runGradientVectorFlow(Image::pointer vectorField) {
        OpenCLDevice::pointer device = getMainDevice();
        reportInfo() << "Running GVF.." << Reporter::end;
        MultigridGradientVectorFlow::pointer gvf = MultigridGradientVectorFlow::New();
        gvf->setInputData(vectorField);
        //gvf->set32bitStorageFormat();
        gvf->set16bitStorageFormat();
        //gvf->setIterations(10);
        //gvf->setMuConstant(0.199);
        //10min
        gvf->setIterations(500);
        gvf->setMuConstant(0.140);
        gvf->update();
        reportInfo() << "GVF finished" << Reporter::end;
        return gvf->getOutputData<Image>();
    }

    Image::pointer CASACLE::createGradients(Image::pointer image) {
        OpenCLDevice::pointer device = getMainDevice();
        Image::pointer floatImage = Image::New();
        floatImage->create(image->getWidth(), image->getHeight(), image->getDepth(), TYPE_FLOAT, 1);
        Image::pointer vectorField = Image::New();
        vectorField->create(image->getWidth(), image->getHeight(), image->getDepth(), TYPE_SNORM_INT16, 3);
        //vectorField->create(image->getWidth(), image->getHeight(), image->getDepth(), TYPE_FLOAT, 3);
        vectorField->setSpacing(image->getSpacing());
        SceneGraph::setParentNode(vectorField, image);

        bool no3Dwrite = !device->isWritingTo3DTexturesSupported();

        OpenCLImageAccess::pointer access = image->getOpenCLImageAccess(ACCESS_READ, device);
        cl::Program program = getOpenCLProgram(device, "", "-DVECTORS_16BIT");
        reportInfo() << "build gradients program" << Reporter::end;

        // Convert to float 0-1
        cl::Kernel toFloatKernel(program, "toFloat");

        float minimumIntensity;
        if (mMinimumIntensity > -std::numeric_limits<float>::max()) {
            minimumIntensity = mMinimumIntensity;
        }
        else {
            minimumIntensity = image->calculateMinimumIntensity();
            std::cout << "min intensity " << minimumIntensity << std::endl;
        }
        float maximumIntensity;
        if (mMaximumIntensity < std::numeric_limits<float>::max()) {
            maximumIntensity = mMaximumIntensity;
        }
        else {
            maximumIntensity = image->calculateMaximumIntensity();
            std::cout << "max intensity " << maximumIntensity << std::endl;
        }
        std::cout << image->getSize().transpose() << std::endl;

        toFloatKernel.setArg(0, *(access->get3DImage()));
        if (no3Dwrite) {
            OpenCLBufferAccess::pointer floatImageAccess = floatImage->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
            toFloatKernel.setArg(1, *(floatImageAccess->get()));
        }
        else {
            OpenCLImageAccess::pointer floatImageAccess = floatImage->getOpenCLImageAccess(ACCESS_READ_WRITE, device);
            toFloatKernel.setArg(1, *(floatImageAccess->get3DImage()));
        }
        toFloatKernel.setArg(2, minimumIntensity);
        toFloatKernel.setArg(3, maximumIntensity);
        device->getCommandQueue().enqueueNDRangeKernel(
            toFloatKernel,
            cl::NullRange,
            cl::NDRange(image->getWidth(), image->getHeight(), image->getDepth()),
            cl::NullRange
            );

        // Create vector field
        //std::cout << "done with to float" << std::endl;
        cl::Kernel vectorFieldKernel(program, "createVectorField");

        // Use sensitivity to set vector maximum (fmax)
        float vectorMaximum = (1 - mSensitivity);
        float sign = mExtractDarkStructures ? -1.0f : 1.0f;

        OpenCLImageAccess::pointer floatImageAccess = floatImage->getOpenCLImageAccess(ACCESS_READ_WRITE, device);
        vectorFieldKernel.setArg(0, *(floatImageAccess->get3DImage()));
        if (no3Dwrite) {
            OpenCLBufferAccess::pointer vectorFieldAccess = vectorField->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
            vectorFieldKernel.setArg(1, *(vectorFieldAccess->get()));
        }
        else {
            OpenCLImageAccess::pointer vectorFieldAccess = vectorField->getOpenCLImageAccess(ACCESS_READ_WRITE, device);
            vectorFieldKernel.setArg(1, *(vectorFieldAccess->get3DImage()));
        }
        vectorFieldKernel.setArg(2, vectorMaximum);
        vectorFieldKernel.setArg(3, sign);
        vectorFieldKernel.setArg(4, vectorField->getSpacing().x());
        vectorFieldKernel.setArg(5, vectorField->getSpacing().y());
        vectorFieldKernel.setArg(6, vectorField->getSpacing().z());
        /*
        vectorFieldKernel.setArg(4, 1.0f);
        vectorFieldKernel.setArg(5, 1.0f);
        vectorFieldKernel.setArg(6, 1.0f);
        */

        // Run kernel
        device->getCommandQueue().enqueueNDRangeKernel(
            vectorFieldKernel,
            cl::NullRange,
            cl::NDRange(image->getWidth(), image->getHeight(), image->getDepth()),
            cl::NullRange
            );

        return vectorField;
    }

    void CASACLE::runTubeDetectionFilter(Image::pointer vectorField, float minimumRadius, float maximumRadius, Image::pointer& TDF, Image::pointer& radius) {
        OpenCLDevice::pointer device = getMainDevice();
        TDF = Image::New();
        TDF->create(vectorField->getSize(), TYPE_FLOAT, 1);
        TDF->setSpacing(vectorField->getSpacing());
        SceneGraph::setParentNode(TDF, vectorField);
        radius = Image::New();
        radius->create(vectorField->getSize(), TYPE_FLOAT, 1);

        OpenCLBufferAccess::pointer TDFAccess = TDF->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
        OpenCLBufferAccess::pointer radiusAccess = radius->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
        OpenCLImageAccess::pointer vectorFieldAccess = vectorField->getOpenCLImageAccess(ACCESS_READ, device);

        cl::Program program = getOpenCLProgram(device);
        cl::Kernel kernel(program, "circleFittingTDF");

        kernel.setArg(0, *(vectorFieldAccess->get3DImage()));
        kernel.setArg(1, *(TDFAccess->get()));
        kernel.setArg(2, minimumRadius);
        kernel.setArg(3, maximumRadius);
        kernel.setArg(4, mRadiusStep);
        kernel.setArg(5, *(radiusAccess->get()));
        /*
        kernel.setArg(6, vectorField->getSpacing().x());
        kernel.setArg(7, vectorField->getSpacing().y());
        kernel.setArg(8, vectorField->getSpacing().z());
        */
        kernel.setArg(6, 1.0f);
        kernel.setArg(7, 1.0f);
        kernel.setArg(8, 1.0f);

        device->getCommandQueue().enqueueNDRangeKernel(
            kernel,
            cl::NullRange,
            cl::NDRange(TDF->getWidth(), TDF->getHeight(), TDF->getDepth()),
            cl::NullRange
            );
    }

    void CASACLE::runCentralTubeDetectionFilter(Image::pointer vectorField, float a, float b, float c, Image::pointer& TDF, Image::pointer& radius, int version, float minimumRadius, float maximumRadius){
        OpenCLDevice::pointer device = getMainDevice();
        TDF = Image::New();
        TDF->create(vectorField->getSize(), TYPE_FLOAT, 1);
        TDF->setSpacing(vectorField->getSpacing());
        SceneGraph::setParentNode(TDF, vectorField);
        radius = Image::New();
        radius->create(vectorField->getSize(), TYPE_FLOAT, 1);

        OpenCLBufferAccess::pointer TDFAccess = TDF->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
        OpenCLBufferAccess::pointer radiusAccess = radius->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
        OpenCLImageAccess::pointer vectorFieldAccess = vectorField->getOpenCLImageAccess(ACCESS_READ, device);

        cl::Program program = getOpenCLProgram(device);
        cl::Kernel kernel(program, "centralTDF");

        kernel.setArg(0, *(vectorFieldAccess->get3DImage()));
        kernel.setArg(1, *(TDFAccess->get()));
        kernel.setArg(2, a);
        kernel.setArg(3, b);
        kernel.setArg(4, c);
        kernel.setArg(5, *(radiusAccess->get()));
        /*
        kernel.setArg(6, vectorField->getSpacing().x());
        kernel.setArg(7, vectorField->getSpacing().y());
        kernel.setArg(8, vectorField->getSpacing().z());
        */
        kernel.setArg(6, 1.0f);
        kernel.setArg(7, 1.0f);
        kernel.setArg(8, 1.0f);
        kernel.setArg(9, version);
        kernel.setArg(10, minimumRadius);
        kernel.setArg(11, maximumRadius);
        kernel.setArg(12, mRadiusStep);

        device->getCommandQueue().enqueueNDRangeKernel(
            kernel,
            cl::NullRange,
            cl::NDRange(TDF->getWidth(), TDF->getHeight(), TDF->getDepth()),
            cl::NullRange
            );
    }

    void CASACLE::runNonCircularTubeDetectionFilter(Image::pointer vectorField, float minimumRadius, float maximumRadius, Image::pointer& TDF, Image::pointer& radius) {
        OpenCLDevice::pointer device = getMainDevice();
        TDF = Image::New();
        TDF->create(vectorField->getSize(), TYPE_FLOAT, 1);
        TDF->setSpacing(vectorField->getSpacing());
        SceneGraph::setParentNode(TDF, vectorField);
        radius = Image::New();
        radius->create(vectorField->getSize(), TYPE_FLOAT, 1);

        OpenCLBufferAccess::pointer TDFAccess = TDF->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
        OpenCLBufferAccess::pointer radiusAccess = radius->getOpenCLBufferAccess(ACCESS_READ_WRITE, device);
        OpenCLImageAccess::pointer vectorFieldAccess = vectorField->getOpenCLImageAccess(ACCESS_READ, device);

        reportInfo() << "building TDF program" << Reporter::end;
        cl::Program program = getOpenCLProgram(device);
        reportInfo() << "build TDF program" << Reporter::end;
        cl::Kernel kernel(program, "nonCircularTDF");

        kernel.setArg(0, *(vectorFieldAccess->get3DImage()));
        kernel.setArg(1, *(TDFAccess->get()));
        kernel.setArg(2, minimumRadius);
        kernel.setArg(3, maximumRadius);
        kernel.setArg(4, mRadiusStep);
        kernel.setArg(5, 12); // nr of line searches
        kernel.setArg(6, 0.2f); // GVF magnitude threshold
        kernel.setArg(7, *(radiusAccess->get()));
        /*
        kernel.setArg(8, vectorField->getSpacing().x());
        kernel.setArg(9, vectorField->getSpacing().y());
        kernel.setArg(10, vectorField->getSpacing().z());
        */
        kernel.setArg(8, 1.0f);
        kernel.setArg(9, 1.0f);
        kernel.setArg(10, 1.0f);

        device->getCommandQueue().enqueueNDRangeKernel(
            kernel,
            cl::NullRange,
            cl::NDRange(TDF->getWidth(), TDF->getHeight(), TDF->getDepth()),
            cl::NullRange
            );
    }

} // end namespace fast
