#include "RTCE.hpp"
#include "FAST/Data/Image.hpp"
#include "FAST/Data/LineSet.hpp"
#include "FAST/Data/Segmentation.hpp"
#include <queue>
#include <vector>
#include <list>
#include <stack>
#include <iostream>
#include <fstream>
#include <boost/unordered_set.hpp>
using boost::unordered_set;
#include <boost/unordered_map.hpp>
using boost::unordered_map;
#define PI 3.14159265;

namespace fast {

RTCE::RTCE() {
    createInputPort<Image>(0);
    createInputPort<Image>(1);
    createInputPort<Image>(2);

    createInputPort<Image>(3);
    createInputPort<Image>(4);
    createInputPort<Image>(5);

    createInputPort<Image>(6);
    // These are not required: Used when centerlines from two TDF results are to be merged
    //createInputPort<Image>(3, false);
    //createInputPort<Image>(4, false);
    //createInputPort<Image>(5, false);

    //createInputPort<Image>(6, false);

    createOutputPort<LineSet>(0, OUTPUT_DEPENDS_ON_INPUT, 0);
    createOutputPort<Segmentation>(1, OUTPUT_DEPENDS_ON_INPUT, 0);
    //createOutputPort<uchar *>(2, OUTPUT_DEPENDS_ON_INPUT, 0);
}

typedef struct point {
    float value;
    int x,y,z;
} point;

typedef struct fPoint {
    float value;
    float x, y, z;
} fPoint;

class PointComparison {
    public:
    bool operator() (const point &lhs, const point &rhs) const {
        return (lhs.value < rhs.value);
    }
};

inline float sign(float a) {
    return a < 0 ? -1.0f: 1.0f;
}

typedef struct CenterlinePoint {
    Vector3i pos;
    bool large;
    Vector3i previousPos;
} CenterlinePoint;

typedef struct SearchPoint{
    point pos;
    point neighbourPos;
} SearchPoint;

#define LPOS(a,b,c) (a)+(b)*(size.x())+(c)*(size.x()*size.y())
#define POS(pos) pos.x()+pos.y()*size.x()+pos.z()*size.x()*size.y()

float squaredMagnitude(ImageAccess::pointer& vectorFieldAccess, Vector3i position) {
    Vector3f vector = vectorFieldAccess->getVector(position).head(3);
    float magnitude = vector.norm();
    return magnitude;
}

float getNormalizedValue(ImageAccess::pointer& vectorField, Vector3i pos, uint component) {
    float magnitude = squaredMagnitude(vectorField, pos);
    if(magnitude == 0) {
        return 0;
    } else {
        return vectorField->getScalar(pos, component) ;/// magnitude;
    }
}

Vector3f gradientNormalized(ImageAccess::pointer& vectorField, Vector3i pos, int volumeComponent, int dimensions) {
    float f100, f_100, f010 = 0, f0_10 = 0, f001 = 0, f00_1 = 0;
    Vector3i npos = pos;
    npos.x() += 1;
    f100 = getNormalizedValue(vectorField, npos, volumeComponent);
    npos.x() -= 2;
    f_100 = getNormalizedValue(vectorField, npos, volumeComponent);
    if(dimensions > 1) {
        npos = pos;
        npos.y() += 1;
        f010 = getNormalizedValue(vectorField, npos, volumeComponent);
        npos.y() -= 2;
        f0_10 = getNormalizedValue(vectorField, npos, volumeComponent);
    }
    if(dimensions > 2) {
        npos = pos;
        npos.z() += 1;
        f001 = getNormalizedValue(vectorField, npos, volumeComponent);
        npos.z() -= 2;
        f00_1 = getNormalizedValue(vectorField, npos, volumeComponent);
    }

    Vector3f grad(0.5f*(f100-f_100), 0.5f*(f010-f0_10), 0.5f*(f001-f00_1));


    return grad;
}

Vector3f gradient(ImageAccess::pointer& vectorField, Vector3i pos, int volumeComponent, int dimensions) {
    float f100, f_100, f010 = 0, f0_10 = 0, f001 = 0, f00_1 = 0;
    Vector3i npos = pos;
    npos.x() += 1;
    f100 = vectorField->getScalar(npos, volumeComponent);
    npos.x() -= 2;
    f_100 = vectorField->getScalar(npos, volumeComponent);
    if(dimensions > 1) {
        npos = pos;
        npos.y() += 1;
        f010 = vectorField->getScalar(npos, volumeComponent);
        npos.y() -= 2;
        f0_10 = vectorField->getScalar(npos, volumeComponent);
    }
    if(dimensions > 2) {
        npos = pos;
        npos.z() += 1;
        f001 = vectorField->getScalar(npos, volumeComponent);
        npos.z() -= 2;
        f00_1 = vectorField->getScalar(npos, volumeComponent);
    }

    Vector3f grad(0.5f*(f100-f_100), 0.5f*(f010-f0_10), 0.5f*(f001-f00_1));


    return grad;
}

void sortEigenvaluesAndVectors(Vector3f* eigenvaluesOut, Matrix3f* eigenvectorsOut) {
    Vector3f eigenvalues = *eigenvaluesOut;
    Matrix3f eigenvectors = *eigenvectorsOut;

    // Find largest eigenvalue
    int largestIndex = 0;
    int smallestIndex = 0;
    for(int i = 1; i < 3; i++) {
        if(fabs(eigenvalues[i]) > fabs(eigenvalues[largestIndex])) {
            largestIndex = i;
        }
        if(fabs(eigenvalues[i]) < fabs(eigenvalues[smallestIndex])) {
            smallestIndex = i;
        }
    }
    Vector3i indexes(0,1,2);
    indexes(smallestIndex) = -1;
    indexes(largestIndex) = -1;
    int middleIndex = indexes.maxCoeff();

    eigenvectorsOut->col(0) = eigenvectors.col(smallestIndex);
    eigenvectorsOut->col(1) = eigenvectors.col(middleIndex);
    eigenvectorsOut->col(2) = eigenvectors.col(largestIndex);
    (*eigenvaluesOut)(0) = eigenvalues(smallestIndex);
    (*eigenvaluesOut)(1) = eigenvalues(middleIndex);
    (*eigenvaluesOut)(2) = eigenvalues(largestIndex);
}


Vector3f getTubeDirection(ImageAccess::pointer& vectorField, Vector3i pos, Vector3ui size, bool normalize) {

    // Do gradient on Fx, Fy and Fz and normalization
    Vector3f Fx, Fy, Fz;
    if(normalize) {
        Fx = gradientNormalized(vectorField, pos,0,1);
        Fy = gradientNormalized(vectorField, pos,1,2);
        Fz = gradientNormalized(vectorField, pos,2,3);
    } else {
        Fx = gradient(vectorField, pos,0,1);
        Fy = gradient(vectorField, pos,1,2);
        Fz = gradient(vectorField, pos,2,3);
    }

    Matrix3f hessian = Matrix3f::Zero();
    hessian(0, 0) = Fx.x();
    hessian(1, 0) = Fy.x();
    hessian(0, 1) = Fy.x();
    hessian(1, 1) = Fy.y();
    hessian.col(2) = Fz;
    hessian.row(2) = Fz;

    Eigen::SelfAdjointEigenSolver<Matrix3f> es(hessian);
    Matrix3f eigenvectors = es.eigenvectors();
    Vector3f eigenvalues = es.eigenvalues();
    sortEigenvaluesAndVectors(&eigenvalues, &eigenvectors);
    return eigenvectors.col(0);
}

void doEigen(ImageAccess::pointer& vectorField, Vector3i pos, Vector3ui size, bool normalize, Vector3f* lambda, Vector3f* e1, Vector3f* e2, Vector3f* e3) {

    // Do gradient on Fx, Fy and Fz and normalization
    Vector3f Fx, Fy, Fz;
    if(normalize) {
        Fx = gradientNormalized(vectorField, pos,0,1);
        Fy = gradientNormalized(vectorField, pos,1,2);
        Fz = gradientNormalized(vectorField, pos,2,3);
    } else {
        Fx = gradient(vectorField, pos,0,1);
        Fy = gradient(vectorField, pos,1,2);
        Fz = gradient(vectorField, pos,2,3);
    }

    Matrix3f hessian = Matrix3f::Zero();
    hessian(0, 0) = Fx.x();
    hessian(1, 0) = Fy.x();
    hessian(0, 1) = Fy.x();
    hessian(1, 1) = Fy.y();
    hessian.col(2) = Fz;
    hessian.row(2) = Fz;

    Eigen::SelfAdjointEigenSolver<Matrix3f> es(hessian);
    Matrix3f eigenvectors = es.eigenvectors();
    Vector3f eigenvalues = es.eigenvalues();
    sortEigenvaluesAndVectors(&eigenvalues, &eigenvectors);
    *lambda = eigenvalues;
    *e1 = eigenvectors.col(0);
    *e2 = eigenvectors.col(1);
    *e3 = eigenvectors.col(2);
}

void copyToLineSet(std::stack<CenterlinePoint> points, std::vector<Vector3f>& vertices, std::vector<Vector2ui>& lines, Vector3f spacing) {
    while(!points.empty()) {
        CenterlinePoint point = points.top();
        points.pop();
        if(point.previousPos.x() != -1) {
            const uint pos = vertices.size();
            vertices.push_back(point.pos.cast<float>().cwiseProduct(spacing));
            vertices.push_back(point.previousPos.cast<float>().cwiseProduct(spacing));
            lines.push_back(Vector2ui(pos, pos+1));
        }
    }
}

/*
void copyToLineSet(std::stack<point> points, std::vector<Vector3f>& vertices, std::vector<Vector2ui>& lines, Vector3f spacing){
    while (!points.empty()){
        point p = points.top();
        points.pop();
        Vector3i pp(p.x, p.y, p.z);
        const uint pos = vertices.size();
        vertices.push_back(pp.cast<float>().cwiseProduct(spacing));
        lines.push_back(Vector2ui(pos, pos + 1));
    }
}*/

void getStartingPoints(
    ImageAccess::pointer& TDFaccess,
    ImageAccess::pointer& vectorFieldAccess,
    float Thigh,
    std::priority_queue<point, std::vector<point>, PointComparison>& queue,
    Vector3ui size,
    Vector3f spacing
    ){

    Vector3i neighborhood[26];
    uint i = 0;
    for (int a = -1; a < 2; a++) {
        for (int b = -1; b < 2; b++) {
            for (int c = -1; c < 2; c++) {
                if (a == 0 && b == 0 && c == 0)
                    continue;
                neighborhood[i] = Vector3i(a, b, c);
                i++;
            }
        }
    }
    
    std::cout << "Getting valid start points for centerline extraction.." << std::endl;
    //float* TDFarray = (float*)TDFaccess->get();
    
    #pragma omp parallel for
    for (int z = 2; z < size.z() - 2; z++) {
        for (int y = 2; y < size.y() - 2; y++) {
            for (int x = 2; x < size.x() - 2; x++) {
                
                //if (TDFarray[x + y*size.x() + z*size.x()*size.y()] < Thigh)
                if (TDFaccess->getScalar(LPOS(x, y, z)) < Thigh){
                    continue;
                }
                Vector3i pos(x, y, z);
                bool valid = true;
                for (int i = 0; i < 26; ++i) {
                    Vector3i nPos = pos + neighborhood[i];
                    if (squaredMagnitude(vectorFieldAccess, nPos) < squaredMagnitude(vectorFieldAccess, pos)) {
                        valid = false;
                        break;
                    }
                }
                
                if (valid) {
                    point p;
                    p.value = TDFaccess->getScalar(LPOS(x, y, z));
                    //p.value = TDFarray[x + y*size.x() + z*size.x()*size.y()];
                    //std::cout << "p value " << p.value << std::endl;
                    p.x = x;
                    p.y = y;
                    p.z = z;
                    #pragma omp critical
                    queue.push(p);
                    
                    if (queue.size() > 100000){
                        std::cout << "queue size " << queue.size() << std::endl;
                    }
                }
            }
        }
    }
    //std::cout << "Counter " << counter << std::endl;
    //delete[] TDFarray;
}

void getStartingPointsDuo(
    ImageAccess::pointer& smallTDFaccess,
    ImageAccess::pointer& smallVFAccess,
    ImageAccess::pointer& largeTDFaccess,
    ImageAccess::pointer& largeVFAccess,
    float largeThigh,
    float smallThigh,
    std::priority_queue<point, std::vector<point>, PointComparison>& queue,
    Vector3ui size,
    Vector3f spacing
    ){

    Vector3i neighborhood[26];
    uint i = 0;
    for (int a = -1; a < 2; a++) {
        for (int b = -1; b < 2; b++) {
            for (int c = -1; c < 2; c++) {
                if (a == 0 && b == 0 && c == 0)
                    continue;
                neighborhood[i] = Vector3i(a, b, c);
                i++;
            }
        }
    }
    
    std::cout << "Getting valid start points for centerline extraction.." << std::endl;
    float* largeTDFarray = (float*)largeTDFaccess->get();
    float* smallTDFarray = (float*)smallTDFaccess->get();
    float largest = 0;
    #pragma omp parallel for
    for (int z = 2; z < size.z() - 2; z++) {
        for (int y = 2; y < size.y() - 2; y++) {
            for (int x = 2; x < size.x() - 2; x++) {

                if (largeTDFarray[x + y*size.x() + z*size.x()*size.y()] < largeThigh && smallTDFarray[x + y*size.x() + z*size.x()*size.y()] < smallThigh){
                    continue;
                }
                //if (TDFaccess->getScalar(LPOS(x, y, z)) < Thigh)
                    

                Vector3i pos(x, y, z);
                bool valid = true;
                for (int i = 0; i < 26; ++i) {
                    Vector3i nPos = pos + neighborhood[i];
                    if (squaredMagnitude(largeVFAccess, nPos) < squaredMagnitude(largeVFAccess, pos) && squaredMagnitude(smallVFAccess, nPos) < squaredMagnitude(smallVFAccess, pos)) {
                        valid = false;
                        break;
                    }
                }

                if (valid) {
                    point p;
                    //p.value = TDFaccess->getScalar(LPOS(x, y, z));
                    p.value = (largeTDFarray[x + y*size.x() + z*size.x()*size.y()] + smallTDFarray[x + y*size.x() + z*size.x()*size.y()])/2;
                    p.x = x;
                    p.y = y;
                    p.z = z;
                    #pragma omp critical
                    queue.push(p);
                }
            }
        }
    }
    
    //delete[] TDFarray;
}

void extractCenterlines(
    Image::pointer TDF,
    Image::pointer vectorField,
    //Image::pointer radius,
    unordered_map<int, int>& distancesMap,
    unordered_map<int, float>& meanMap,
    int* centerlines,
    int* centerlinesInfo,
    bool* radiusInfo,
    bool radiusOrder,
    int maxBelowTlow,
    float Thigh,
    float Tlow,
    //float Mlow,
    int lMin,
    int& counter,
    float meanMin,
    float Mlow,
    unordered_map<int, std::vector<CenterlinePoint>>& centerlineMap
    ){
    
    ImageAccess::pointer tdfAccess = TDF->getImageAccess(ACCESS_READ);
    ImageAccess::pointer vectorFieldAccess = vectorField->getImageAccess(ACCESS_READ);
    //ImageAccess::pointer radiusAccess = radius->getImageAccess(ACCESS_READ);

    const Vector3ui size = TDF->getSize();
    const Vector3f spacing = vectorField->getSpacing();
    const int totalSize = size.x()*size.y()*size.z();
    
    int distance = 0;
    bool queueCheck = false;
    std::priority_queue<point, std::vector<point>, PointComparison> queue;
    
    while (queue.size() == 0 || queue.size() > 800 || queue.size() < 300){
        
        getStartingPoints(tdfAccess, vectorFieldAccess, Thigh, queue, size, spacing);
        std::cout << "Queue size " << queue.size() << " Tlow " << Tlow << " Thigh " << Thigh << std::endl;
        if (queue.size() < 300){
            if (Tlow < 0.015){
                Tlow = Tlow - 0.0005;
            }
            else{
                Tlow = Tlow - 0.005;
            }
            Thigh = Thigh - 0.01;
            queue = std::priority_queue<point, std::vector<point>, PointComparison>();
        }
        else if (queue.size() > 800){
            if (Tlow < 0.015){
                Tlow = Tlow + 0.0005;
            }
            else{
                Tlow = Tlow + 0.005;
            }
            Thigh = Thigh + 0.01;
            queue = std::priority_queue<point, std::vector<point>, PointComparison>();
        }
    }
    std::cout << "Thigh " << Thigh << " Tlow " << Tlow << std::endl;
    std::cout << "Processing " << queue.size() << " valid start points" << std::endl;
    if (queue.size() == 0) {
        //throw Exception("no valid start points found");
        queueCheck = true;
    }
    
    
    while (!queue.empty()) {
        // Traverse from new start point
        point p = queue.top();
        queue.pop();

        // Has it been handled before?
        if (centerlines[LPOS(p.x, p.y, p.z)] != 0)
            continue;
        
        Vector3i prevPoint(p.x, p.y, p.z);
        Vector3i prevprevPoint(p.x, p.y, p.z);
        
        //Because output:(
        std::vector<CenterlinePoint> currentCenterline;
        CenterlinePoint clp;
        clp.previousPos = Vector3i(-1, -1, -1);
        clp.pos = Vector3i(p.x, p.y, p.z);
        clp.large = radiusOrder;
        currentCenterline.insert(currentCenterline.begin(), clp);
        
        
        int distance = 1;
        
        //Se etter flere steder der meantube skal vre
        float meanTube = tdfAccess->getScalar(clp.pos);
        //float meanTube = p.value;
        //Create a new vector for this centerline
        std::vector<point> cpVector;

        //Add first point
        cpVector.insert(cpVector.begin(), p);


        // For each direction
        for (int direction = -1; direction < 3; direction += 2) {
            int belowTlow = 0;
            Vector3i position(p.x, p.y, p.z);
            prevPoint = Vector3i(p.x-1, p.y-1, p.z-1);
            prevprevPoint = Vector3i(p.x-2, p.y-2, p.z-2);
            Vector3f t_i = getTubeDirection(vectorFieldAccess, position, size, maxBelowTlow > 0)*direction;
            Vector3f t_i_1 = t_i;


            // Traverse
            while (true) {

                Vector3i maxPoint(0, 0, 0);

                // Check for out of bounds
                if (position.x() < 3 || position.x() > size.x() - 3 || position.y() < 3 || position.y() > size.y() - 3 || position.z() < 3 || position.z() > size.z() - 3)
                    break;

                // Try to find next point from all neighbors
                for (int a = -1; a < 2; a++) {
                    for (int b = -1; b < 2; b++) {
                        for (int c = -1; c < 2; c++) {
                            Vector3i n(position.x() + a, position.y() + b, position.z() + c);
                            if ((a == 0 && b == 0 && c == 0))
                                continue;

                            Vector3f dir = (n - position).cast<float>();//.cwiseProduct(spacing);
                            dir.normalize();
                            if (dir.dot(t_i) <= 0.1) // Maintain direction
                                continue;

                            // Is magnitude smaller than previous
                            if (maxPoint == Vector3i(0, 0, 0)) {
                                maxPoint = n;
                            }
                            else if (1 - squaredMagnitude(vectorFieldAccess, n) > 1 - squaredMagnitude(vectorFieldAccess, maxPoint)) {
                                maxPoint = n;
                            }
                        }
                    }
                }

                if (maxPoint.x() + maxPoint.y() + maxPoint.z() > 0) {
                    //Finne ut om den har truffet en existerende centerline
                    //Finne ut om maxpoint er godt nok

                    if (centerlines[POS(maxPoint)] != 0){
                        bool loopDetected = false;
                        for (int j = 0; j < cpVector.size(); j++){
                            if (cpVector[j].x == maxPoint.x() && cpVector[j].y == maxPoint.y() && cpVector[j].z == maxPoint.z()){
                                //Loop detected
                                loopDetected = true;
                                
                            }
                        }
                        if (loopDetected){
                            //might need to erase the centerline here
                            for (int i = 0; i < cpVector.size(); i++){
                                Vector3i rePoint(cpVector[i].x, cpVector[i].y, cpVector[i].z);
                                centerlines[POS(rePoint)] = 0;
                            }
                            currentCenterline.clear();
                            distancesMap[counter] = 0;
                            meanMap[counter] = 0;
                            //break;
                        }
                        else{
                            int connectionValue = centerlines[POS(maxPoint)];
                            distancesMap[connectionValue] = distance;
                            meanMap[connectionValue] = meanTube;
                            for (int i = 0; i < cpVector.size(); i++){
                                Vector3i rePoint(cpVector[i].x, cpVector[i].y, cpVector[i].z);
                                centerlines[POS(rePoint)] = connectionValue;
                            }
                            
                            for (int j = 0; j < currentCenterline.size(); j++){
                                centerlineMap[connectionValue].insert(centerlineMap[connectionValue].begin(), currentCenterline[j]);
                            }
                            meanMap[connectionValue] += meanTube;
                            distancesMap[connectionValue] += distance;
                            distancesMap[counter] = 0;
                            meanMap[counter] = 0;
                            currentCenterline.clear();
                        }
                        break;
                    }
                    //else if (tdfAccess->getScalar(maxPoint) < Tlow){

                    //else if (belowTlow > maxBelowTlow && tdfAccess->getScalar(maxPoint) < Tlow){
                    else if (1 - squaredMagnitude(vectorFieldAccess, maxPoint) < Mlow || (belowTlow > maxBelowTlow && tdfAccess->getScalar(maxPoint) < Tlow)){
                        //Vet ikke om jeg skal ha Mlow og belowTlow her
                        //Point is to low
                        //Add last point to  endpoint here
                        centerlinesInfo[POS(prevPoint)] = counter;
                        centerlinesInfo[POS(prevprevPoint)] = -counter;
                        radiusInfo[POS(prevprevPoint)] = radiusOrder;
                        radiusInfo[POS(prevPoint)] = radiusOrder;
                        distancesMap[counter] = distance;
                        meanMap[counter] = meanTube;

                        centerlineMap[counter] = currentCenterline;
                        break;
                    }
                    else{
                        //std::cout << "Scalar " << tdfAccess->getScalar(maxPoint) << std::endl;
                        if (tdfAccess->getScalar(maxPoint) < Tlow) {
                            belowTlow++;
                        }
                        else {
                            belowTlow = 0;
                        }

                        Vector3f lambda, e1, e2, e3;
                        doEigen(vectorFieldAccess, maxPoint, size, maxBelowTlow > 0, &lambda, &e1, &e2, &e3);
                        if ((lambda.x() < 0 && lambda.y() < 0 && lambda.z() < 0)) {
                            if (fabs(t_i.dot(e3)) > fabs(t_i.dot(e2))) {
                                if (fabs(t_i.dot(e3)) > fabs(t_i.dot(e1))) {
                                    e1 = e3;
                                }
                            }
                            else if (fabs(t_i.dot(e2)) > fabs(t_i.dot(e1))) {
                                e1 = e2;
                            }
                        }

                        float maintain_dir = sign(e1.dot(t_i));
                        Vector3f vec_sum;
                        vec_sum.x() = maintain_dir*e1.x() + t_i.x() + t_i_1.x();
                        vec_sum.y() = maintain_dir*e1.y() + t_i.y() + t_i_1.y();
                        vec_sum.z() = maintain_dir*e1.z() + t_i.z() + t_i_1.z();
                        vec_sum.normalize();
                        t_i_1 = t_i;
                        t_i = vec_sum;

                        float newScalar = tdfAccess->getScalar(maxPoint);
                        meanTube += newScalar;
                        point newPoint;
                        //newPoint.value = newScalar;
                        newPoint.value = counter;
                        newPoint.x = maxPoint.x();
                        newPoint.y = maxPoint.y();
                        newPoint.z = maxPoint.z();

                        CenterlinePoint newCLPoint;
                        newCLPoint.pos = maxPoint;
                        newCLPoint.previousPos = prevPoint;
                        newCLPoint.large = radiusOrder;
                        currentCenterline.insert(currentCenterline.begin(), newCLPoint);

                        //Point is okey?
                        centerlines[POS(maxPoint)] = counter;
                        cpVector.insert(cpVector.begin(), newPoint);
                        radiusInfo[POS(maxPoint)] = radiusOrder;
                        distance++;
                        prevprevPoint = prevPoint;
                        prevPoint = maxPoint;
                        position = maxPoint;
                        //std::cout << "Traversing" << std::endl;
                    }
                }
                else{
                    if (prevPoint.x() + prevPoint.y() + prevPoint.z() > 0){
                        centerlinesInfo[POS(prevPoint)] = counter;
                        centerlinesInfo[POS(prevprevPoint)] = -counter;
                        radiusInfo[POS(prevPoint)] = radiusOrder;
                        radiusInfo[POS(prevprevPoint)] = radiusOrder;
                        distancesMap[counter] = distance;
                        meanMap[counter] = meanTube;
                        std::cout << "No maxPoint, but there is prevPoint" << std::endl;
                    }
                    else{
                        std::cout << "No maxPoint, and no prevPoint, something went wrong" << std::endl;
                    }
                    //No maxpoint
                    break;
                }
            }
        }
        centerlineMap[counter] = currentCenterline;
        counter++;
        //std::cout << "counter " << counter << std::endl;
    }
    
    //Done with queue
    //Traverse and remove unwanted centerlines
    //Then call grouping
    std::cout << "Done with queue" << std::endl;
    std::cout << "Removing to short centerlines" << std::endl;
    std::vector<int> linesToBeRemoved;
    if (!queueCheck){
        for (int i = 0; i < distancesMap.size(); i++){
            if (distancesMap[i] < lMin){
                linesToBeRemoved.insert(linesToBeRemoved.begin(), i);
                distancesMap[i] = 0;
            }
        }
    }
    
    //Maybe I need meanTube stuff here
    /*
    for (int m = 0; m < meanMap.size(); m++){
        if (meanMap[m] < meanMin){
            linesToBeRemoved.insert(linesToBeRemoved.begin(), m);
            meanMap[m] = 0;
        }
    }*/
    std::cout << "Lines to be removed " << linesToBeRemoved.size() << std::endl;
    if (!linesToBeRemoved.empty()){
        #pragma omp parallel for
        for (int x = 2; x < size.x() - 2; x++) {
            for (int y = 2; y < size.y() - 2; y++) {
                for (int z = 2; z < size.z() - 2; z++) {
                    int tCLV = abs(centerlines[LPOS(x, y, z)]);
                    int tEPV = abs(centerlinesInfo[LPOS(x, y, z)]);
                    if (tCLV != 0 || tEPV != 0){
                        for (int j = 0; j < linesToBeRemoved.size(); j++){
                            if (tCLV == linesToBeRemoved[j]){
                                centerlines[LPOS(x, y, z)] = 0;
                            }
                            if (tEPV == linesToBeRemoved[j]){
                                centerlinesInfo[LPOS(x, y, z)] = 0;
                            }
                        }
                    }
                }
            }
        }
        for (int k = 0; k < linesToBeRemoved.size(); k++){
            centerlineMap[linesToBeRemoved[k]].clear();
        }
    }
    centerlines[LPOS(0, 0, 0)] = 0;
    centerlinesInfo[LPOS(0, 0, 0)] = 0;
}

void performGrouping(
    Image::pointer origImg,
    int* centerlines,
    int* centerlinesInfo,
    float angleVar,
    float grayDiff,
    float connectionMax,
    int searchDistance,
    unordered_map<int, std::vector<CenterlinePoint> >& centerlineMap,
    bool* radiusInfo,
    ImageAccess::pointer& largeRadiusAccess,
    ImageAccess::pointer& smallRadiusAccess
    ){
    
    const Vector3ui size = origImg->getSize();
    //const Vector3f spacing = vectorField->getSpacing();
    const int totalSize = size.x()*size.y()*size.z();
    ImageAccess::pointer origImgAccess = origImg->getImageAccess(ACCESS_READ);
    std::cout << "Performing grouping" << std::endl;
    std::vector<point> endPoints;
    std::vector<point> prevPoints;
    std::vector<SearchPoint> searchPoints;
    for (int x0 = 0; x0 < size.x(); x0++){
        for (int y0 = 0; y0 < size.y(); y0++){
            for (int z0 = 0; z0 < size.z(); z0++){
                int tV = centerlinesInfo[LPOS(x0, y0, z0)];
                if (tV == 0){
                    continue;
                }
                if (tV > 1 || tV == 1){
                    point p;
                    p.value = tV;
                    p.x = x0; p.y = y0; p.z = z0;
                    endPoints.insert(endPoints.begin(), p);
                }
                if (tV < -1 || tV == -1){
                    point p;
                    p.value = tV;
                    p.x = x0; p.y = y0; p.z = z0;
                    prevPoints.insert(prevPoints.begin(), p);
                }
            }
        }
    }
    if (prevPoints.size() != endPoints.size()){
        //something went wrong, some endpoints/prevpoints doesnt have corresponding point
        std::cout << "some endpoints/prevpoints doesnt have corresponding point" << std::endl;
    }
    std::cout << "done finding endpoints" << std::endl;
    for (int i = 0; i < endPoints.size(); i++){
        int tempCounter = 0;
        for (int j = 0; j < prevPoints.size(); j++){
            if (abs(endPoints[i].value) == abs(prevPoints[j].value)){
                SearchPoint tS;
                tS.pos = endPoints[i];
                tS.neighbourPos = prevPoints[j];
                searchPoints.insert(searchPoints.begin(), tS);
                tempCounter++;
            }
        }
        if (tempCounter > 2){
            //something has gone wrong, more then one matching point
        }
    }
    //SearchPoints should have info now
    if (searchPoints.empty()){
        //something went wrong
    }
    std::cout << "done matching points" << std::endl;
    for (int k = 0; k < searchPoints.size(); k++){
        //find bestTarget
        point bestTarget;
        float bestScore = 0.0;
        float modD1 = sqrt(abs((searchPoints[k].neighbourPos.x - searchPoints[k].pos.x)*(searchPoints[k].neighbourPos.x - searchPoints[k].pos.x)
            + (searchPoints[k].neighbourPos.y - searchPoints[k].pos.y)*(searchPoints[k].neighbourPos.y - searchPoints[k].pos.y)
            + (searchPoints[k].neighbourPos.z - searchPoints[k].pos.z)*(searchPoints[k].neighbourPos.z - searchPoints[k].pos.z)));
        bestTarget.value = 0; bestTarget.x = 0; bestTarget.y = 0; bestTarget.z = 0;
        for (int x = searchPoints[k].pos.x - searchDistance; x < searchPoints[k].pos.x + searchDistance; x++){
            for (int y = searchPoints[k].pos.y - searchDistance; y < searchPoints[k].pos.y + searchDistance; y++){
                for (int z = searchPoints[k].pos.z - searchDistance; z < searchPoints[k].pos.z + searchDistance; z++){
                    if (x < 20 || x > size.x() - 20 || y < 20 || y > size.y() - 20 || z < 20 || z > size.z() - 20){
                        continue;
                    }
                    if (centerlines[LPOS(x, y, z)] != 0 && centerlines[LPOS(x, y, z)] != searchPoints[k].pos.value){
                        Vector3i endPos(searchPoints[k].pos.x, searchPoints[k].pos.y, searchPoints[k].pos.z);
                        Vector3i targetPos(x,y,z);
                        float grayChange = abs(origImgAccess->getScalar(endPos) - origImgAccess->getScalar(targetPos));
                        
                        if (grayChange > grayDiff){
                            continue;
                        }
                        float targetValue = 0;

                        //Manual angle calc, change later
                        float dotProd = (searchPoints[k].neighbourPos.x - endPos.x())*(endPos.x() - targetPos.x())
                            + (searchPoints[k].neighbourPos.y - endPos.y())*(endPos.y() - targetPos.y())
                            + (searchPoints[k].neighbourPos.z - endPos.z())*(endPos.z() - targetPos.z());
                        float modD2 = sqrt(abs((endPos.x() - targetPos.x())*(endPos.x() - targetPos.x())
                            + (endPos.y() - targetPos.y())*(endPos.y() - targetPos.y())
                            + (endPos.z() - targetPos.z())*(endPos.z() - targetPos.z())));
                        float cosnum = dotProd / (modD1*modD2);
                        float cos = acos(cosnum) * 180 / PI;
                        float angle = abs(1 - (cos / 180));
                        targetValue = modD2 / exp(-angle / (2 * (angleVar*angleVar)));
                        if (targetValue > bestScore && targetValue < connectionMax){
                            bestTarget.value = centerlines[LPOS(x, y, z)];
                            bestTarget.x = targetPos.x(); bestTarget.y = targetPos.y(); bestTarget.z = targetPos.z();
                            //targetValue = bestScore;
                            bestScore = targetValue;
                        }
                        /*
                        if (targetValue > bestTarget.value && targetValue < connectionMax){
                            bestTarget.value = targetValue;
                            bestTarget.x = targetPos.x(); bestTarget.y = targetPos.y(); bestTarget.z = targetPos.z();
                        }*/
                    }
                }
            }
        }
        if (bestTarget.value != 0){
            //Perform traversal
            //std::cout << "bestTarget found" << std::endl;
            bool targetReached = false;
            Vector3i currentPoint(searchPoints[k].pos.x, searchPoints[k].pos.y, searchPoints[k].pos.z);
            int prevX = searchPoints[k].neighbourPos.x; int prevY = searchPoints[k].neighbourPos.y; int prevZ = searchPoints[k].neighbourPos.z;
            while (!targetReached){
                //Vector3i bestNewPoint(0, 0, 0);
                int bestX = 0; int bestY = 0; int bestZ = 0;
                
                float interDistance = 99.0;
                 
                for (int x = currentPoint.x() - 1; x < currentPoint.x() + 2; x++){
                    for (int y = currentPoint.y() - 1; y < currentPoint.y() + 2; y++){
                        for (int z = currentPoint.z() - 1; z < currentPoint.z() + 2; z++){
                            float tempDistance = sqrt(abs((x - bestTarget.x)*(x - bestTarget.x)
                                + (y - bestTarget.y)*(y - bestTarget.y)
                                + (z - bestTarget.z)*(z - bestTarget.z)));
                            if (tempDistance < interDistance || (x == bestTarget.x && y == bestTarget.y && z == bestTarget.z)){
                                bestX = x; bestY = y; bestZ = z;
                                interDistance = tempDistance;
                            }
                        }
                    }
                }
                if (bestX == bestTarget.x && bestY == bestTarget.y && bestZ == bestTarget.z){
                    targetReached = true;
                    
                    for (int i = 0; i < centerlineMap[searchPoints[k].pos.value].size(); i++){
                        centerlineMap[bestTarget.value].insert(centerlineMap[bestTarget.value].begin(), centerlineMap[searchPoints[k].pos.value][i]);
                        //centerlines[LPOS(searchPoints[k].pos.x,searchPoints[k].pos.y,searchPoints[k].pos.z)] = bestTarget.value;
                        centerlines[POS(centerlineMap[searchPoints[k].pos.value][i].pos)] = bestTarget.value;
                        if (radiusInfo[POS(centerlineMap[searchPoints[k].pos.value][i].pos)] != true){
                            uchar largeTemp = round(largeRadiusAccess->getScalar(centerlineMap[searchPoints[k].pos.value][i].pos));
                            uchar smallTemp = round(smallRadiusAccess->getScalar(centerlineMap[searchPoints[k].pos.value][i].pos));
                            if (largeTemp > smallTemp){
                                radiusInfo[POS(centerlineMap[searchPoints[k].pos.value][i].pos)] = true;
                            }
                            else if (largeTemp == 0 && smallTemp == 0){
                                radiusInfo[POS(centerlineMap[searchPoints[k].pos.value][i].pos)] = false;
                            }
                            else{
                                radiusInfo[POS(centerlineMap[searchPoints[k].pos.value][i].pos)] = false;
                            }
                        }
                        //if (largeRadiusAccess->getScalar())
                        /*
                        if (radiusInfo[LPOS(searchPoints[k].pos.x, searchPoints[k].pos.y, searchPoints[k].pos.z)] != true){
                            if (largeRadiusAccess->getScalar(Vector3i(searchPoints[k].pos) > smallRadiusAccess->getScalar(searchPoints[k].pos)){

                            }
                        }*/
                        //radiusInfo[LPOS(searchPoints[k].pos.x, searchPoints[k].pos.y, searchPoints[k].pos.z)] = true;
                    }
                    centerlineMap[searchPoints[k].pos.value].clear();
                    break;
                }
                else if (bestX + bestY + bestZ != 0){
                    centerlines[LPOS(bestX, bestY, bestZ)] = bestTarget.value;
                    radiusInfo[LPOS(bestX, bestY, bestZ)] = true;
                    //currentPoint.x = bestX; currentPoint.y = bestY; currentPoint.z = bestZ;
                    currentPoint = Vector3i(bestX,bestY,bestZ);
                    CenterlinePoint newPoint;
                    newPoint.pos = currentPoint;
                    newPoint.previousPos = Vector3i(prevX, prevY, prevZ);
                    newPoint.large = true;
                    centerlineMap[bestTarget.value].insert(centerlineMap[bestTarget.value].begin(), newPoint);

                    uchar largeTemp = largeRadiusAccess->getScalar(newPoint.pos);
                    uchar smallTemp = smallRadiusAccess->getScalar(newPoint.pos);
                    if (radiusInfo[POS(newPoint.pos)] != true){
                        if (largeTemp > smallTemp){
                            radiusInfo[POS(newPoint.pos)] = true;
                        }
                        else if (largeTemp == 0 && smallTemp == 0){
                            radiusInfo[POS(newPoint.pos)] = false;
                        }
                        else{
                            radiusInfo[POS(newPoint.pos)] = false;
                        }
                    }
                    prevX = bestX; prevY = bestY; prevZ = bestZ;
                    //bestX = 0; bestY = 0; bestZ = 0;
                    //interDistance = 0;
                }
                else{
                    //Something went wrong
                    std::cout << "Something went wrong" << std::endl;
                }
            }
        }
        else{
            //no besttarget
            //std::cout << "No besttarget" << std::endl;
        }
    }
    std::cout << "done with grouping" << std::endl;

}

void extractCenterlinesTEMP(
        Image::pointer TDF,
        Image::pointer vectorField,
        Image::pointer radius,
        int* centerlines,
        unordered_map<int, int>& centerlineDistances,
        unordered_map<int, std::stack<CenterlinePoint> >& centerlineStacks,
        std::vector<Vector3f>& vertices,
        std::vector<Vector2ui>& lines,
        int maxBelowTlow,
        bool* useFirstRadius,
        unordered_map<int, std::vector<CenterlinePoint> >& centerlineVectors,
        unordered_map<int, std::vector<CenterlinePoint> >& endPointMap
    ) {
    ImageAccess::pointer TDFaccess = TDF->getImageAccess(ACCESS_READ);
    ImageAccess::pointer vectorFieldAccess = vectorField->getImageAccess(ACCESS_READ);
    const Vector3ui size = TDF->getSize();
    const Vector3f spacing = vectorField->getSpacing();

    static int counter = 1;
    //float Thigh = 0.5;
    float Thigh = 1.0;
    int Dmin = 10;//getParam(parameters, "min-distance");
    //float Mlow = 0.1;
    //float Tlow = 0.1;
    //float Mlow = 0.000004;
    float Mlow = 0.1;
    //float Tlow = 0.000004;
    float Tlow = 0.5;
    float minMeanTube = maxBelowTlow > 0 ? 0.4 : 0.5;
    //float minMeanTube = 0.0;
    const int totalSize = size.x()*size.y()*size.z();

    Vector3i neighborhood[26];
    uint i = 0;
    for(int a = -1; a < 2; a++) {
        for(int b = -1; b < 2; b++) {
            for(int c = -1; c < 2; c++) {
                if(a == 0 && b == 0 && c == 0)
                    continue;
                neighborhood[i] = Vector3i(a,b,c);
                i++;
            }
        }
    }

        // Create queue
    std::priority_queue<point, std::vector<point>, PointComparison> queue;

    std::cout << "Getting valid start points for centerline extraction.." << std::endl;
    float* TDFarray = (float*)TDFaccess->get();
    /*
    // Collect all valid start points
    int counter99 = 0;
    
    int ovCounter = 0;
    float ovAvg = 0;
    int counter0 = 0;
    int counter1 = 0;
    float ovAvg1 = 0;
    int counter2 = 0;
    int counter3 = 0;
    int counter4 = 0;
    int counter5 = 0;
    int counter6 = 0;
    int counter10 = 0;
    
    
    #pragma omp parallel for
    for (int z = 2; z < size.z() - 2; z++) {
        for (int y = 2; y < size.y() - 2; y++) {
            for (int x = 2; x < size.x() - 2; x++) {
                
                if (TDFarray[x + y*size.x() + z*size.x()*size.y()] < Thigh)
                    //if(TDFaccess->getScalar(Vector3i(x,y,z)) < Thigh) // This is reaaaaallllly slow
                    continue;
                
                float tdfValue = TDFarray[x + y*size.x() + z*size.x()*size.y()];
                //Mellom 0.00001 - 0.000001
                //0.000005 - 9k
                //0.0000075 - 0
                //0.000007 - 0
                //0.000006 - 1415
                //0.0000065 - 173
                if (tdfValue < Thigh && tdfValue > Tlow){
                    counter99++;
                    //std::cout << "TDFvalue: " << tdfValue << std::endl;
                }
                
                Vector3i pos(x, y, z);
                //float tdfValue = TDFaccess->getScalar(Vector3i(x, y, z));
                //tdfValue = roundf((tdfValue * 100) / 100);
                //std::cout << "TDFvalue: " << tdfValue << std::endl;
                //if (tdfValue > 0.1){
                  //  std::cout << "TDFvalue: " << tdfValue << std::endl;
                //}
                //if (tdfValue)
                //#pragma omp critical
                
                ovCounter += 1;
                //#pragma omp critical
                ovAvg += tdfValue;
                if (tdfValue > 0 && tdfValue < 1){
                    //#pragma omp critical
                    counter0 += 1;
                    //#pragma omp critical
                    ovAvg1 += tdfValue;
                    //std::cout << "ScalarValue: " << tdfValue << std::endl;
                }
                else if (tdfValue > 1 && tdfValue < 2){
                    //#pragma omp critical
                    counter1 += 1;
                }
                else if (tdfValue > 2 && tdfValue < 3){
                    //#pragma omp critical
                    counter2 += 1;
                }
                else if (tdfValue > 3 && tdfValue < 4){
                    //#pragma omp critical
                    counter3 += 1;
                }
                else if (tdfValue > 4 && tdfValue < 5){
                    //#pragma omp critical
                    counter4 += 1;
                }
                else if (tdfValue > 5 && tdfValue < 6){
                    //#pragma omp critical
                    counter5 += 1;
                }
                else if (tdfValue > 6 && tdfValue < 7){
                    //#pragma omp critical
                    counter6 += 1;
                }
                else if (tdfValue > 10){
                    //#pragma omp critical
                    counter10 += 1;
                }
                 
                    
                
                
                if (tdfValue > Tlow){
                    point p;
                    p.value = tdfValue;
                    p.x = x;
                    p.y = y;
                    p.z = z;
                    #pragma omp critical
                    queue.push(p);
                    //std::cout << "pushing point: " << ovCounter << std::endl;
                }
            }
        }
    }
    std::cout << "pushing point: " << counter99 << std::endl;
    
    std::cout << "ovCounter: " << ovCounter << std::endl;
    std::cout << "ovAvg: " << ovAvg/ovCounter << std::endl;
    std::cout << "counter1: " << counter1 << std::endl;
    std::cout << "ovAvg1: " << ovAvg1/counter1 << std::endl;
    std::cout << "counter2: " << counter2 << std::endl;
    std::cout << "counter3: " << counter3 << std::endl;
    std::cout << "counter4: " << counter4 << std::endl;
    std::cout << "counter5: " << counter5 << std::endl;
    std::cout << "counter6: " << counter6 << std::endl;
    std::cout << "counter10: " << counter10 << std::endl;
    */
    
    #pragma omp parallel for
    for(int z = 2; z < size.z()-2; z++) {
        for(int y = 2; y < size.y()-2; y++) {
            for(int x = 2; x < size.x()-2; x++) {
                if(TDFarray[x + y*size.x() + z*size.x()*size.y()] < Thigh)
                //if(TDFaccess->getScalar(Vector3i(x,y,z)) < Thigh) // This is reaaaaallllly slow
                    continue;

                Vector3i pos(x,y,z);
                bool valid = true;
                for(int i = 0; i < 26; ++i) {
                    Vector3i nPos = pos + neighborhood[i];
                    if(squaredMagnitude(vectorFieldAccess, nPos) < squaredMagnitude(vectorFieldAccess, pos)) {
                        valid = false;
                        break;
                    }
                }

                if(valid) {
                    point p;
                    p.value = TDFaccess->getScalar(Vector3i(x,y,z));
                    p.x = x;
                    p.y = y;
                    p.z = z;
                    #pragma omp critical
                    queue.push(p);
                }
            }
        }
    }
    

    std::cout << "Processing " << queue.size() << " valid start points" << std::endl;
    if(queue.size() == 0) {
        throw Exception("no valid start points found");
    }

    while(!queue.empty()) {
        // Traverse from new start point
        point p = queue.top();
        queue.pop();

        // Has it been handled before?
        if(centerlines[LPOS(p.x,p.y,p.z)] == 1)
            continue;

        unordered_set<int> newCenterlines;
        newCenterlines.insert(LPOS(p.x,p.y,p.z));
        int distance = 1;
        int connections = 0;
        int prevConnection = -1;
        int secondConnection = -1;
        float meanTube = TDFaccess->getScalar(Vector3i(p.x,p.y,p.z));

        //Create a new vector for this centerline Endpoints
        std::vector<CenterlinePoint> endPoints;
        //Create a new vector for this centerline for grouping
        std::vector<CenterlinePoint> cpVector;

        // Create new stack for this centerline
        std::stack<CenterlinePoint> stack;
        CenterlinePoint startPoint;
        startPoint.previousPos = Vector3i(-1, -1, -1);
        startPoint.pos.x() = p.x;
        startPoint.pos.y() = p.y;
        startPoint.pos.z() = p.z;
        
        stack.push(startPoint);
        cpVector.insert(cpVector.begin(), startPoint);

        // For each direction
        for(int direction = -1; direction < 3; direction += 2) {
            Vector3i previous = startPoint.pos;
            int belowTlow = 0;
            Vector3i position(p.x,p.y,p.z);
            Vector3f t_i = getTubeDirection(vectorFieldAccess, position, size, maxBelowTlow > 0)*direction;
            Vector3f t_i_1 = t_i;


            // Traverse
            while(true) {
                Vector3i maxPoint(0,0,0);

                // Check for out of bounds
                if(position.x() < 3 || position.x() > size.x()-3 || position.y() < 3 || position.y() > size.y()-3 || position.z() < 3 || position.z() > size.z()-3)
                    break;

                // Try to find next point from all neighbors
                for(int a = -1; a < 2; a++) {
                    for(int b = -1; b < 2; b++) {
                        for(int c = -1; c < 2; c++) {
                            Vector3i n(position.x()+a,position.y()+b,position.z()+c);
                            if((a == 0 && b == 0 && c == 0))
                                continue;

                            Vector3f dir = (n - position).cast<float>();//.cwiseProduct(spacing);
                            dir.normalize();
                            if(dir.dot(t_i) <= 0.1) // Maintain direction
                                continue;

                            //if(T.radius[POS(n)] >= 1.5f) {
                            // Is magnitude smaller than previous
                            if(maxPoint == Vector3i(0,0,0)) {
                                maxPoint = n;
                            } else if(1 - squaredMagnitude(vectorFieldAccess, n) > 1 - squaredMagnitude(vectorFieldAccess, maxPoint)) {
                                maxPoint = n;
                            }
                            /*
                            } else {
                                if(TDFaccess->getScalar(n)*(1-squaredMagnitude(vectorFieldAccess, n)) > TDFaccess->getScalar(maxPoint)*(1-squaredMagnitude(vectorFieldAccess, maxPoint)))
                                maxPoint = n;
                            }
                            */

                        }
                    }
                }

                if(maxPoint.x() + maxPoint.y() + maxPoint.z() > 0) {
                    // New maxpoint found, check it!
                    if(centerlines[POS(maxPoint)] > 0) {
                        // Hit an existing centerline
                        if(prevConnection == -1) {
                            prevConnection = centerlines[POS(maxPoint)];
                            // Add connection point
                            CenterlinePoint p;
                            p.pos = maxPoint;
                            p.previousPos = previous;
                            previous = position;
                            stack.push(p);
                            
                            //Added
                            cpVector.insert(cpVector.begin(), p);
                            
                            distance ++;
                            newCenterlines.insert(POS(maxPoint));
                            meanTube += TDFaccess->getScalar(maxPoint);
                        } else {
                            if(prevConnection == centerlines[POS(maxPoint)]) {
                                // A loop has occured, reject this centerline
                                connections = 5;
                            } else {
                                secondConnection = centerlines[POS(maxPoint)];
                                // Add connection point
                                CenterlinePoint p;
                                p.pos = maxPoint;
                                p.previousPos = previous;
                                previous = position;
                                stack.push(p);
                                //Added
                                cpVector.insert(cpVector.begin(), p);
                                distance ++;
                                newCenterlines.insert(POS(maxPoint));
                                meanTube += TDFaccess->getScalar(maxPoint);
                            }
                        }
                        break;
                    } else if(1 - squaredMagnitude(vectorFieldAccess, maxPoint) < Mlow || (belowTlow > maxBelowTlow && TDFaccess->getScalar(maxPoint) < Tlow)) {
                        // New point is below thresholds
                        break;
                    } else if(newCenterlines.count(POS(maxPoint)) > 0) {
                        // Loop detected!
                        break;
                    } else {
                        // Point is OK, proceed to add it and continue
                        if(TDFaccess->getScalar(maxPoint) < Tlow) {
                            belowTlow++;
                        } else {
                            belowTlow = 0;
                        }

                        // Update direction
                        //float3 e1 = getTubeDirection(T, maxPoint,size.x,size.y,size.z);

                        //TODO: check if all eigenvalues are negative, if so find the egeinvector that best matches
                        Vector3f lambda, e1, e2, e3;
                        doEigen(vectorFieldAccess, maxPoint, size, maxBelowTlow > 0, &lambda, &e1, &e2, &e3);
                        if((lambda.x() < 0 && lambda.y() < 0 && lambda.z() < 0)) {
                            if(fabs(t_i.dot(e3)) > fabs(t_i.dot(e2))) {
                                if(fabs(t_i.dot(e3)) > fabs(t_i.dot(e1))) {
                                    e1 = e3;
                                }
                            } else if(fabs(t_i.dot(e2)) > fabs(t_i.dot(e1))) {
                                e1 = e2;
                            }
                        }

                        float maintain_dir = sign(e1.dot(t_i));
                        Vector3f vec_sum;
                        vec_sum.x() = maintain_dir*e1.x() + t_i.x() + t_i_1.x();
                        vec_sum.y() = maintain_dir*e1.y() + t_i.y() + t_i_1.y();
                        vec_sum.z() = maintain_dir*e1.z() + t_i.z() + t_i_1.z();
                        vec_sum.normalize();
                        t_i_1 = t_i;
                        t_i = vec_sum;

                        // update position
                        position = maxPoint;
                        distance ++;
                        newCenterlines.insert(POS(maxPoint));
                        meanTube += TDFaccess->getScalar(maxPoint);

                        // Create centerline point
                        CenterlinePoint p;
                        p.pos = position;
                        p.previousPos = previous;
                        previous = position;
                        /*
                        if(T.radius[POS(p.pos)] > 3.0f) {
                            p.large = true;
                        } else {
                            p.large = false;
                        }
                        */

                        // Add point to stack
                        stack.push(p);

                        //Added
                        cpVector.insert(cpVector.begin(), p);
                    }
                } else {
                    // No maxpoint found, stop!
                    // Add endpoint here i think
                    break;
                }

            } // End traversal
        } // End for each direction

        // Check to see if new traversal can be added
        std::cout << "Finished. Distance " << distance << " meanTube: " << meanTube/distance << std::endl;
        if(distance > Dmin && meanTube/distance > minMeanTube && connections < 2) {
            //std::cout << "Finished. Distance " << distance << " meanTube: " << meanTube/distance << std::endl;
            //std::cout << "------------------- New centerlines added #" << counter << " -------------------------" << std::endl;


            unordered_set<int>::iterator usit;
            if(prevConnection == -1) {
                // No connections
                for(usit = newCenterlines.begin(); usit != newCenterlines.end(); usit++) {
                    centerlines[*usit] = counter;
                    if(maxBelowTlow > 0) {
                        useFirstRadius[*usit] = true;
                    }
                }
                centerlineDistances[counter] = distance;
                centerlineStacks[counter] = stack;
                //Added
                centerlineVectors[counter] = cpVector;
                counter ++;
            } else {
                // The first connection
                // TODO connect together in stack

                std::stack<CenterlinePoint> prevConnectionStack = centerlineStacks[prevConnection];
                //Added
                std::vector<CenterlinePoint> prevConnectionVector;
                while(!stack.empty()) {
                    prevConnectionStack.push(stack.top());
                    stack.pop();

                    //Added
                    int tempCounter = 0;
                    prevConnectionVector.insert(prevConnectionVector.begin(), cpVector[tempCounter]);
                    tempCounter++;
                }
                //Added
                cpVector.clear();

                for(usit = newCenterlines.begin(); usit != newCenterlines.end(); usit++) {
                    centerlines[*usit] = prevConnection;
                    if(maxBelowTlow > 0) {
                        useFirstRadius[*usit] = true;
                    }
                }
                centerlineDistances[prevConnection] += distance;
                if(secondConnection != -1) {
                    // Two connections, move secondConnection to prevConnection
                    std::stack<CenterlinePoint> secondConnectionStack = centerlineStacks[secondConnection];
                    //Added
                    std::vector<CenterlinePoint> secondConnectionVector = centerlineVectors[secondConnection];
                    centerlineVectors.erase(secondConnection);

                    centerlineStacks.erase(secondConnection);
                    while(!secondConnectionStack.empty()) {
                        prevConnectionStack.push(secondConnectionStack.top());
                        secondConnectionStack.pop();

                        int tempCounter2 = 0;
                        prevConnectionVector.insert(prevConnectionVector.begin(),secondConnectionVector[tempCounter2]);
                        tempCounter2++;
                    }

                    #pragma omp parallel for
                    for(int i = 0; i < totalSize;i++) {
                        if(centerlines[i] == secondConnection)
                            centerlines[i] = prevConnection;
                    }
                    centerlineDistances[prevConnection] += centerlineDistances[secondConnection];
                    centerlineDistances.erase(secondConnection);
                }

                centerlineStacks[prevConnection] = prevConnectionStack;
                centerlineVectors[prevConnection] = prevConnectionVector;
            }
        } // end if new point can be added
    } // End while queue is not empty
    std::cout << "Finished traversal" << std::endl;

    //tror centerlineVectors er hele greia n....
    //If needed, the grouping can be done here as well
}

bool sortBySize(const std::vector<CenterlinePoint> &first, const std::vector<CenterlinePoint> &second){ return first.size() > second.size(); }

void RTCE::execute(){
    //Output
    LineSet::pointer centerlineOutput = getStaticOutputData<LineSet>(0);
    Segmentation::pointer centerlineVolumeOutput = getStaticOutputData<Segmentation>(1);


    Image::pointer largeTDF = getStaticInputData<Image>(0);
    Image::pointer vectorField = getStaticInputData<Image>(1);
    Image::pointer largeRadius = getStaticInputData<Image>(2);
    Image::pointer smallTDF = getStaticInputData<Image>(3);
    Image::pointer gradients = getStaticInputData<Image>(4);
    Image::pointer smallRadius = getStaticInputData<Image>(5);
    Image::pointer origInputImage = getStaticInputData<Image>(6);
    

    const Vector3ui size = largeTDF->getSize();
    const Vector3f spacing = vectorField->getSpacing();
    const int totalSize = size.x()*size.y()*size.z();

    int * centerlines = new int[totalSize]();
    int * centerlinesInfo = new int[totalSize]();
    bool * radiusInfo = new bool[totalSize]();
    unordered_map<int, int > distancesMap;
    unordered_map<int, float > meanMap;
    unordered_map<int, std::vector<CenterlinePoint> > centerlineMap;

    std::vector<Vector3f> vertices;
    std::vector<Vector2ui> lines;

    bool radiusOrder = true; //LargeRadius = true, false = small
    int maxBelowTlow = 12;
    float Thigh = 0.11;//0.85;
    //float Tlow = 0.04;//0.02;
    float Tlow = 0.04;//0.06;//0.02;
    float Mlow = 0.1;
    float Lmin = 3;
    int counter = 0;
    float meanMin = 100;

    float angleVar = 0.7;
    float grayDiff = 200;
    float connectionMax = 20;
    int searchDistance = 20;

    extractCenterlines(largeTDF,vectorField,distancesMap,meanMap,centerlines,centerlinesInfo,radiusInfo,radiusOrder,maxBelowTlow,Thigh,Tlow,Lmin,counter,meanMin, Mlow,centerlineMap);
    if (getNrOfInputData() > 3){
        radiusOrder = false;
        maxBelowTlow = 4;
        Tlow = 0.008;//0.01;
        Thigh = 0.09;//0.91;
        //Tlow = 0.4;
        
        extractCenterlines(smallTDF, gradients, distancesMap, meanMap, centerlines, centerlinesInfo, radiusInfo, radiusOrder, maxBelowTlow, Thigh, Tlow, Lmin, counter, meanMin, Mlow,centerlineMap);
    }
    ImageAccess::pointer largeRadiusAccess = largeRadius->getImageAccess(ACCESS_READ);
    ImageAccess::pointer smallRadiusAccess = smallRadius->getImageAccess(ACCESS_READ);
    performGrouping(origInputImage,centerlines,centerlinesInfo,angleVar,grayDiff,connectionMax,searchDistance,centerlineMap, radiusInfo,largeRadiusAccess,smallRadiusAccess);
    
    unordered_map<int, std::vector<CenterlinePoint> > tempVectorMap;
    unordered_map<int, std::stack<CenterlinePoint> > stackMap;
    int tempCounter = 0;
    std::cout << "centerlineMap " << centerlineMap.size() << std::endl;
    for (int i = 0; i < centerlineMap.size(); i++){
        if (!centerlineMap[i].empty()){
            tempVectorMap[tempCounter] = centerlineMap[i];
            tempCounter++;
        }
    }
    std::vector<std::vector<CenterlinePoint>> sortedVector;
    for (int i0 = 0; i0 < tempVectorMap.size(); i0++){
        sortedVector.insert(sortedVector.begin(), tempVectorMap[i0]);
    }
    //Bare legge ved, etc 10 lengste
    //Sjekke om lengden p centerlinemap og centerlines er lik
    std::sort(sortedVector.begin(),sortedVector.end(),sortBySize);
    std::cout << "Sorting done " << std::endl;
    int numberOfLines = 0;
    uchar * returnCenterlines = new uchar[totalSize]();
    for (int fill = 0; fill < totalSize; fill++){
        returnCenterlines[fill] = 0;
    }
    std::cout << "before while" << std::endl;
    while (true){
        if (sortedVector[numberOfLines].size() < 20 || sortedVector.size() - 3 < numberOfLines){
            break;
        }
        numberOfLines++;
    }
    std::cout << "after while" << std::endl;
    if (numberOfLines > sortedVector.size() - 1){
        numberOfLines = sortedVector.size();
    }
    numberOfLines = 2;
    //numberOfLines = sortedVector.size()-1;
    std::cout << "Lines displayed " << numberOfLines << std::endl;
    std::ofstream clTxt;
    std::ofstream refTxt;
    clTxt.open("tubeguicenterline.txt");
    refTxt.open("tubeguireference.txt");
    ImageAccess::pointer largeAccess = largeRadius->getImageAccess(ACCESS_READ);
    ImageAccess::pointer smallAccess = smallRadius->getImageAccess(ACCESS_READ);
    for (int m = 0; m < numberOfLines; m++){
        for (int n = 0; n < sortedVector[m].size(); n++){
            Vector3i tempPoint = sortedVector[m][n].pos;
            float radius = 0;
            if (radiusInfo[POS(tempPoint)]){
                radius = largeAccess->getScalar(tempPoint);
            }
            else{
                radius = smallAccess->getScalar(tempPoint);
            }
            if (radius < 0.5 || radius == 0){
                radius = 1.0;
            }
            clTxt << tempPoint.x() << " " << tempPoint.y() << " " << tempPoint.z() << std::endl;
            refTxt << tempPoint.x() << " " << tempPoint.y() << " " << tempPoint.z() << " " << radius << " 0.0" << std::endl;
        }
    }
    clTxt.close();
    refTxt.close();
    largeAccess->release();
    smallAccess->release();

    for (int j = 0; j < numberOfLines; j++){
        std::stack<CenterlinePoint> tempStack;
        for (int k = 0; k < sortedVector[j].size(); k++){
            tempStack.push(sortedVector[j][k]);
            if (radiusInfo[POS(sortedVector[j][k].pos)] == true){
                uchar radi = round(largeRadiusAccess->getScalar(sortedVector[j][k].pos));
                if (radi == 0 || radi == 1){
                    returnCenterlines[POS(sortedVector[j][k].pos)] = 1;
                }
                else{
                    returnCenterlines[POS(sortedVector[j][k].pos)] = radi;
                }
            }
            else{
                uchar radi = round(smallRadiusAccess->getScalar(sortedVector[j][k].pos));
                if (radi == 0 || radi == 1){
                    returnCenterlines[POS(sortedVector[j][k].pos)] = 1;
                }
                else{
                    returnCenterlines[POS(sortedVector[j][k].pos)] = radi;
                }
            }
        }
        stackMap[j] = tempStack;
    }
    
    std::cout << "Making lineset" << std::endl;
    for (int l = 0; l < stackMap.size(); l++){
        copyToLineSet(stackMap[l], vertices, lines, spacing);
    }
    std::cout << "done with lineset" << std::endl;

    
    
    //std::cout << "Creating output 1" << std::endl;
    centerlineOutput->create(vertices, lines);
    //std::cout << "Creating output 2" << std::endl;
    centerlineVolumeOutput->create(size.x(), size.y(), size.z(), TYPE_UINT8, 1, getMainDevice(), returnCenterlines);
    //std::cout << "Creating output 3" << std::endl;
    //setStaticOutputData<uchar>(2, returnCenterlines);
    delete[] returnCenterlines;
    delete[] centerlines;
    delete[] centerlinesInfo;
    delete[] radiusInfo;
    centerlineVolumeOutput->setSpacing(largeTDF->getSpacing());
    //std::cout << "Creating output 4" << std::endl;
    SceneGraph::setParentNode(centerlineVolumeOutput, largeTDF);
    //std::cout << "Creating output 5" << std::endl;
    SceneGraph::setParentNode(centerlineOutput, largeTDF);
    //std::cout << "Creating output 6" << std::endl;
}

void RTCE::TEMPexecute() {
    std::cout << "RTCE started" << std::endl;
    LineSet::pointer centerlineOutput = getStaticOutputData<LineSet>(0);
    Segmentation::pointer centerlineVolumeOutput = getStaticOutputData<Segmentation>(1);

    Image::pointer TDF = getStaticInputData<Image>(0);

    //Orig input
    Image::pointer origInputImage = getStaticInputData<Image>(6);
    ImageAccess::pointer origInputAccess = origInputImage->getImageAccess(ACCESS_READ);

    Vector3ui size = TDF->getSize();
    const int totalSize = size.x()*size.y()*size.z();
    int TreeMin = 10;

    // Create some data structures
    int * centerlines = new int[totalSize]();

    // Create a map of centerline distances
    unordered_map<int, int> centerlineDistances;

    // Create a map of centerline stacks
    unordered_map<int, std::stack<CenterlinePoint> > centerlineStacks;

    // Create a map of centerline vectors for grouping
    unordered_map<int, std::vector<CenterlinePoint> > centerlineVectors;

    // Create a map of centerline endpoints per centerline
    unordered_map<int, std::vector<CenterlinePoint> > endPointMap;

    std::vector<Vector3f> vertices;
    std::vector<Vector2ui> lines;

    uint firstLimit;
    int counter = 1;
    bool* useFirstRadius = new bool[totalSize]();
    Image::pointer radius = getStaticInputData<Image>(2);
    {
        Image::pointer vectorField = getStaticInputData<Image>(1);
        extractCenterlinesTEMP(TDF, vectorField, radius, centerlines, centerlineDistances, centerlineStacks, vertices, lines, 12, useFirstRadius, centerlineVectors, endPointMap);
        // TODO do inverse gradient segmentation here?
    }

    // Check to see if more than two inputs were provided, if so run again..
    Image::pointer radius2;
    if(getNrOfInputData() > 3) {
        // Run again for small
        Image::pointer TDF = getStaticInputData<Image>(3);
        Image::pointer vectorField = getStaticInputData<Image>(4);
        radius2 = getStaticInputData<Image>(5);
        extractCenterlinesTEMP(TDF, vectorField, radius2, centerlines, centerlineDistances, centerlineStacks, vertices, lines, 0, useFirstRadius, centerlineVectors, endPointMap);

        // TODO do dilation segmentation here?
    }

    

    if(centerlineDistances.size() == 0) {
        //throw SIPL::SIPLException("no centerlines were extracted");
        std::cout << "No centerlines were extracted" << std::endl;
        delete[] centerlines;
        return;
    }
    std::cout << centerlineDistances.size() << " centerline extracted" << std::endl;

    // TODO do grouping/endpoint search here?
    //centerlineVectors skal n ha alle punktene, kan hende at jeg ikke trenger stacks
    //Tror det beste er og ta en og en vector, dvs, frst finne endpoints i en vector, s gjre sk per point, 
    //for vrt point s gjr jeg connection, trenger en eller annen oversikt, kanskje lagre connection point, 
    //kombinere nr alle points er skt ferdig
    //Jeg trenger kanskje en liste over vectors to be erased...
    int searchDistance = 20;
    float connectionMax = 200;
    float angleVar = 0.7;
    float grayDiff = 200;
    //ImageAccess::pointer TDFaccess = TDF->getImageAccess(ACCESS_READ);
    //float* TDFarray = (float*)TDFaccess->get();

    std::cout << "Performing Grouping/Linking" << std::endl;
    for (int g1 = 0; g1 < centerlineVectors.size(); g1++){
        std::vector<CenterlinePoint> currentVector = centerlineVectors[g1];
        
        //Find enpoints, this might be a stupid way to do this...
        std::vector<CenterlinePoint> endPoints;
        for (int e1 = 0; e1 < currentVector.size(); e1++){
            bool endPointCheck = false;
            Vector3i v1 = currentVector[e1].pos;
            for (int e2 = 0; e2 < currentVector.size(); e2++){
                Vector3i v2 = currentVector[e2].previousPos;
                if (v1 == v2){
                    endPointCheck = true;
                    //break;
                }
            }
            if (endPointCheck){
                endPointCheck = false;
            }
            else{
                std::cout << "Found endpoint!" << std::endl;
                endPoints.insert(endPoints.begin(), currentVector[e1]);
                endPointCheck = false;
            }
        }
        if (endPoints.empty()){
            std::cout << "No endpoints found" << std::endl;
            continue;
        }
        for (int s1 = 0; s1 < endPoints.size(); s1++){
            std::cout << "Performing linking of individual" << std::endl;
            std::vector<Vector3i> searchSpace;
            CenterlinePoint cEP = endPoints[s1];
            //Find potential targets
            for (int x1 = endPoints[s1].pos.x()-searchDistance; x1 < endPoints[s1].pos.x()+searchDistance; x1++){
                for (int y1 = endPoints[s1].pos.y()-searchDistance; y1 < endPoints[s1].pos.y()+searchDistance; y1++){
                    for (int z1 = endPoints[s1].pos.z()-searchDistance; z1 < endPoints[s1].pos.z()+searchDistance; z1++){
                        //Add out of bounds here
                        if (x1-2 > size.x() || y1-2 > size.y() || z1-2 > size.z() || x1 < 2 || y1 < 2 || z1 < 2){
                            continue;
                        }
                        Vector3i checkPoint(x1, y1, z1);
                        if (centerlines[POS(checkPoint)] > 0){
                            bool isInCenterline = false;
                            for (int clC = 0; clC < centerlineVectors[g1].size(); clC++){
                                if (centerlineVectors[g1][clC].pos.x() == checkPoint.x() && centerlineVectors[g1][clC].pos.y() == checkPoint.y() && centerlineVectors[g1][clC].pos.z() == checkPoint.z()){
                                    isInCenterline = true;
                                }
                            }
                            if (!isInCenterline){
                                searchSpace.insert(searchSpace.begin(), checkPoint);
                            }
                            //std::cout << "Found intersecting centerlines" << std::endl;
                        }
                    }
                }
            }
            if (searchSpace.empty()){
                std::cout << "No intersecting centerlines" << std::endl;
                continue;
            }
            //Linepoints is endPoints[s1].pos and endPoints[s1].previousPos, where .pos is the endpoint
            //searchSpace[s2] is the targetPoint
            std::cout << "Finding best target" << std::endl;
            Vector3i bestTarget(0, 0, 0);
            float bestValue = 0;
            //std::cout << "modD1 start" << std::endl;
            float modD1 = sqrt(abs((cEP.previousPos.x() - cEP.pos.x())*(cEP.previousPos.x() - cEP.pos.x())
                + (cEP.previousPos.y() - cEP.pos.y())*(cEP.previousPos.y() - cEP.pos.y())
                + (cEP.previousPos.z() - cEP.pos.z())*(cEP.previousPos.z() - cEP.pos.z())));
            //float modD1 = sqrt((cEP.previousPos.x() - cEP.pos.x())*(cEP.previousPos.x() - cEP.pos.x())
              //  + (cEP.previousPos.y() - cEP.pos.y())*(cEP.previousPos.y() - cEP.pos.y())
                //+ (cEP.previousPos.z() - cEP.pos.z())*(cEP.previousPos.z() - cEP.pos.z()));
            //std::cout << "modD1 end" << std::endl;
            //Find best target
            for (int s2 = 0; s2 < searchSpace.size(); s2++){
                //This does not work atm
                //float grayChange = abs(TDFaccess->getScalar(cEP.pos) - TDFaccess->getScalar(searchSpace[s2]));
                //float grayChange = abs(TDFarray[cEP.pos.x() + cEP.pos.y()*size.x() + cEP.pos.z()*size.x()*size.y()] - TDFarray[searchSpace[s2].x() + searchSpace[s2].y()*size.x() + searchSpace[s2].z()*size.x()*size.y()]);
                //origInputImage->
                float grayChange = abs(origInputAccess->getScalar(cEP.pos) - origInputAccess->getScalar(searchSpace[s2]));
                //TDFarray[x + y*size.x() + z*size.x()*size.y()]
                //TDFarray[searchSpace[s2].x() + searchSpace[s2].y()*size.x() + searchSpace[s2].z()*size.x()*size.y()]
                //std::cout <<  "grayChange: " << grayChange << std::endl;
                //For testing
                if (grayChange > grayDiff){
                    continue;
                }
                float targetValue = 0;
                //Manual angle calc, change later
                float dotProd = (cEP.previousPos.x()-cEP.pos.x())*(cEP.pos.x()-searchSpace[s2].x())
                    + (cEP.previousPos.y()-cEP.pos.y())*(cEP.pos.y()-searchSpace[s2].y())
                    + (cEP.previousPos.z()-cEP.pos.z())*(cEP.pos.z()-searchSpace[s2].z());
                float modD2 = sqrt((cEP.pos.x() - searchSpace[s2].x())*(cEP.pos.x() - searchSpace[s2].x())
                    + (cEP.pos.y() - searchSpace[s2].y())*(cEP.pos.y() - searchSpace[s2].y())
                    + (cEP.pos.z() - searchSpace[s2].z())*(cEP.pos.z() - searchSpace[s2].z()));
                float cosnum = dotProd/(modD1*modD2);
                //float cos = acos(cosnum)*180/PI;
                float cos = acos(cosnum)*180/PI;
                //std::cout << "angle " << cos << std::endl;
                float angle = abs(1 - (cos / 180));
                //float distance = (searchSpace[s2].x() - cEP.pos.x())*(searchSpace[s2].x() - cEP.pos.x());
                targetValue = modD2 / exp(-angle / (2 * (angleVar*angleVar)));
                //if (targetValue > connectionMax){
                  //  continue;
                //}
                //Score formula
                //(TDFaccess->getScalar(Vector3i(x, y, z)) < Thigh)
                //std::cout << "TargetValue " << targetValue << std::endl;
                if (targetValue > bestValue && targetValue < connectionMax){
                    bestValue = targetValue;
                    bestTarget = searchSpace[s2];
                }
            }
            //Score check
            //if (bestValue == 0){
                //std::cout << "bestValue to low: " << bestValue << std::endl;
              //  continue;
            //}
            std::cout << "Starting search" << std::endl;
            bool targetReached = false;
            //Vector3i currentTarget(bestTarget.x(),bestTarget.y(),bestTarget.z());
            Vector3i currentTarget(cEP.pos.x(), cEP.pos.y(), cEP.pos.z());
            
            if (bestTarget.x() + bestTarget.y() + bestTarget.z() > 1){
                float lowestDistance = 99;
                //Target found
                //Perform search, either linear interpolation or simple a*
                //Super simple search, find neighbours, if a neighbour is closer than orig point, add it and continue
                while (!targetReached){
                    //Vector3i bestAlt(0, 0, 0);
                    int bestAltX = 0; int bestAltY = 0; int bestAltZ = 0;
                    //float lowestDistance = 99;
                    for (int a1 = currentTarget.x() - 1; a1 <= currentTarget.x() + 1; a1++){
                        for (int a2 = currentTarget.y() - 1; a2 <= currentTarget.y() + 1; a2++){
                            for (int a3 = currentTarget.z() - 1; a3 <= currentTarget.z() + 1; a3++){
                                //dont need OOB here, because of search
                                float tempDistance = sqrt(abs((currentTarget.x() - a1)*(currentTarget.x() - a1)
                                    + (currentTarget.y() - a2)*(currentTarget.y() - a2)
                                    + (currentTarget.z() - a3)*(currentTarget.z() - a3)));
                                //std::cout << "Distance: " << tempDistance << std::endl;
                                if (a1 == bestTarget.x() && a2 == bestTarget.y() && a3 == bestTarget.z()){
                                    bestAltX = a1; bestAltY = a2; bestAltZ = a3;
                                    lowestDistance = -1;
                                }
                                else if (tempDistance < lowestDistance || tempDistance == 0){
                                    lowestDistance = tempDistance;
                                    bestAltX = a1; bestAltY = a2; bestAltZ = a3;
                                }
                                /*
                                if (tempDistance < lowestDistance){
                                    lowestDistance = tempDistance;
                                    bestAltX = a1; bestAltY = a2; bestAltZ = a3;
                                }
                                if (a1 == bestTarget.x() && a2 == bestTarget.y() && a3 == bestTarget.z()){
                                    bestAltX = a1; bestAltY = a2; bestAltZ = a3;
                                    lowestDistance = -1;
                                    //std::cout << "continue in search" << std::endl;
                                    //continue;
                                }*/
                            }
                        }
                    }
                    if (lowestDistance == 99){
                        std::cout << "Something whet wrong in search" << std::endl;
                        targetReached = true;
                        //break;
                    }
                    //Here i need to add to centerlineStack + centerlineVector
                    
                    
                    Vector3i bestAlt(bestAltX, bestAltY, bestAltZ);
                    if (lowestDistance == -1 || lowestDistance == 0){
                        targetReached = true;
                        break;
                    }
                    else{
                        CenterlinePoint newCenterlinePoint;
                        newCenterlinePoint.previousPos = currentTarget;
                        newCenterlinePoint.pos = bestAlt;
                        //Need to input into centerlines here
                        centerlines[POS(bestAlt)] = g1;
                        centerlineStacks[g1].push(newCenterlinePoint);
                        centerlineVectors[g1].insert(centerlineVectors[g1].begin(), newCenterlinePoint);
                        centerlineDistances[g1] + 1;
                        currentTarget = bestAlt;
                        std::cout << "Point created" << std::endl;
                    }
                    /*
                    //I have noe idea if this is working
                    //centerlineStacks[g1].push(newCenterlinePoint);
                    //currentVector.insert(currentVector.begin(), newCenterlinePoint);
                    //It never finds the bestTarget for some reason
                    if (bestAlt.x() == bestTarget.x() && bestAlt.y() == bestTarget.y() && bestAlt.z() == bestTarget.z()){
                        targetReached = true;
                        //std::cout << "Target reached" << std::endl;
                    }else{
                        
                        CenterlinePoint newCenterlinePoint;
                        newCenterlinePoint.previousPos = currentTarget;
                        newCenterlinePoint.pos = bestAlt;
                        centerlineStacks[g1].push(newCenterlinePoint);
                        centerlineVectors[g1].insert(centerlineVectors[g1].begin(), newCenterlinePoint);
                        centerlineDistances[g1] + 1;
                        currentTarget = bestAlt;
                        //std::cout << "Point created" << std::endl;
                    }*/
                    
                    //Need to do something with centerlines here, dont know what yet
                }
            }
        }
    }
    std::cout << "Finished grouping/linkage" << std::endl;

    std::cout << "Finding largest tree" << std::endl;
    // Find largest connected tree and all trees above a certain size
    unordered_map<int, int>::iterator it;
    int max = centerlineDistances.begin()->first;
    std::list<int> trees;
    for(it = centerlineDistances.begin(); it != centerlineDistances.end(); it++) {
        if(it->second > centerlineDistances[max])
            max = it->first;
        if(it->second > TreeMin)
            trees.push_back(it->first);
    }
    std::list<int>::iterator it2;
    // TODO: if use the method with TreeMin have to add them to centerlineStack also
    std::stack<CenterlinePoint> centerlineStack = centerlineStacks[max];
    for(it2 = trees.begin(); it2 != trees.end(); it2++) {
        copyToLineSet(centerlineStacks[*it2], vertices, lines, TDF->getSpacing());
        while(!centerlineStacks[*it2].empty()) {
            centerlineStack.push(centerlineStacks[*it2].top());
            centerlineStacks[*it2].pop();
        }
    }

    std::ofstream rTxt;
    rTxt.open("radius.txt");

    std::cout << "Marking largest tree" << std::endl;
    //uchar * returnCenterlines = getStaticOutputData<uchar *>(2);
    uchar * returnCenterlines = new uchar[totalSize]();
    ImageAccess::pointer radiusAccess = radius->getImageAccess(ACCESS_READ);
    ImageAccess::pointer radius2Access;
    if(radius2.isValid())
        radius2Access = radius2->getImageAccess(ACCESS_READ);
    // Mark largest tree with 1, and rest with 0
    /*
    #pragma omp parallel for
    for(int i = 0; i < totalSize;i++) {
        bool valid = false;
        std::list<int>::iterator it2;
        for(it2 = trees.begin(); it2 != trees.end(); it2++) {
            if(centerlines[i] == *it2) {
                // Store radius in centerline volume
                if(useFirstRadius[i]) {
                    returnCenterlines[i] = round(radiusAccess->getScalar(i));
                    rTxt << "Large: " << radiusAccess->getScalar(i) << std::endl;
                } else {
                    returnCenterlines[i] = 1;//round(radius2Access->getScalar(i));
                    rTxt << "Small: " << radius2Access->getScalar(i) << std::endl;
                }
                valid = true;
                break;
            }
        }
        if(!valid)
            returnCenterlines[i] = 0;
    }
    std::cout << "Done marking" << std::endl;
    delete[] centerlines;
    */
    //Lines
    for (int lc = 0; lc < lines.size(); lc++){
        //std::cout << "Line: " << lc << ": " << lines[lc] << std::endl;
    }


    //Txt
    std::ofstream clTxt;
    std::ofstream refTxt;
    
    clTxt.open("centerline.txt");
    refTxt.open("reference.txt");

    for (int txt = 0; txt < vertices.size(); txt++){
        Vector3f tempPoint = vertices[txt];
        Vector3i rPoint(tempPoint.x(),tempPoint.y(),tempPoint.z());
        clTxt << tempPoint.x() << " " << tempPoint.y() << " " << tempPoint.z() << std::endl;
        //refTxt << tempPoint.x() << " " << tempPoint.y() << " " << tempPoint.z() << std::endl;
        refTxt << tempPoint.x() << " " << tempPoint.y() << " " << tempPoint.z() << " 1.5 " << "0.0" << std::endl;
        //rTxt << radiusAccess->getScalar(rPoint) << " " << radius2Access->getScalar(rPoint) << std::endl;

    }
    clTxt.close();
    refTxt.close();
    rTxt.close();


    //std::cout << "Creating output 1" << std::endl;
    centerlineOutput->create(vertices, lines);
    //std::cout << "Creating output 2" << std::endl;
    centerlineVolumeOutput->create(size.x(), size.y(), size.z(), TYPE_UINT8, 1, getMainDevice(), returnCenterlines);
    //std::cout << "Creating output 3" << std::endl;
    //setStaticOutputData<uchar>(2, returnCenterlines);
    delete[] returnCenterlines;
    centerlineVolumeOutput->setSpacing(TDF->getSpacing());
    //std::cout << "Creating output 4" << std::endl;
    SceneGraph::setParentNode(centerlineVolumeOutput, TDF);
    //std::cout << "Creating output 5" << std::endl;
    SceneGraph::setParentNode(centerlineOutput, TDF);
    //std::cout << "Creating output 6" << std::endl;
}

}
